"""
Argument value parsing

This module is mostly self-contained, and provides ways to construct :class:`.Parser` instances
which parse string arguments into values.

During the configuration building, the parsed values are collected by a
:class:`configpile.collector.Collector` instance.

.. rubric:: Types

This module uses the following types.

.. py:data:: _Value

    Value being parsed by a :class:`.Parser`

.. py:data:: _Parameter

    Type received by a mapping function

.. py:data:: _ReturnType

    Type returned by a mapping function

.. py:data:: _Item

    Item in a sequence
"""

from __future__ import annotations

import pathlib
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Callable,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import Protocol, runtime_checkable

from .enums import ForceCase
from .userr import Err, Res, collect_seq
from .util import assert_never

if TYPE_CHECKING:
    try:
        import parsy
    except ImportError:
        pass

# those types are documented in the module docstring

_I = TypeVar("_I")  # type used only internally

_Value = TypeVar("_Value")

_Value_contra = TypeVar("_Value_contra", contravariant=True)

_Item = TypeVar("_Item")

_Parameter = TypeVar("_Parameter")

_ReturnType = TypeVar("_ReturnType")


@runtime_checkable
class Predicate(Protocol, Generic[_Value_contra]):
    """
    Defines a function or callable that takes a value and returns whether it is valid
    """

    def __call__(self, value: _Value_contra) -> bool:
        """
        Checks whether the given value is valid

        Args:
            value: Value to check

        Returns:
            True if the value is valid, False otherwise
        """


@runtime_checkable
class ErrorMessageProvider(Protocol, Generic[_Value_contra]):
    """
    Generates formatted error messages when values are invalid
    """

    def __call__(self, string_to_parse: str, parsed_value: _Value_contra) -> str:
        """
        Returns a formatted error message describing a validation problem

        Args:
            string_to_parse: String provided by the user
            parsed_value: Value parsed from string
        """


class Parser(ABC, Generic[_Value]):
    """Describes a parser, that takes a string and returns either a value or an error"""

    @abstractmethod
    def parse(self, arg: str) -> Res[_Value]:
        """
        Parses a string into a result

        This method reports parsing errors using a result type instead of raising
        exceptions.

        Example:
            >>> user_input = "invalid"
            >>> float_parser.parse(user_input)
            Err1(msg="...", contexts=[])
            >>> user_input = 2.0
            >>> float_parser.parse(user_input)
            2.0

        Args:
            arg: String value to parse

        Returns:
            A result containing either the parsed value or a description of an error
        """

    def choices(self) -> Optional[Sequence[str]]:
        """
        Returns, when relevant, a set of inputs whose parse results cover all possible values

        The definition above is written carefully: parsers may parse aliases, or normalize input
        when parsing (by stripping whitespace or normalizing case).

        This method returns a set of strings, which does not need to be minimal, so that
        their parsed values cover the set of possible result values.
        """
        return None

    def map(self, f: Callable[[_Value], _Item]) -> Parser[_Item]:
        """
        Maps successful results of the parser through the given function

        Args:
            f: Function to map the result through

        Returns:
            Updated parser
        """
        return _Mapped(self, f)

    def flat_map(self, f: Callable[[_Value], Res[_Item]]) -> Parser[_Item]:
        """
        Maps successful results of the parser through the given function that may fail

        Args:
            f: Function to map the result through, may return an error

        Returns:
            Updated parser
        """
        return _FlatMapped(self, f)

    def as_sequence_of_one(self) -> Parser[Sequence[_Value]]:
        """
        Returns a parser, that returns a sequence of a single value on success

        Returns:
            Updated parser
        """
        f: Callable[[_I], Sequence[_I]] = lambda t: [t]
        return self.map(f)

    def empty_means_none(self, strip: bool = True) -> Parser[Optional[_Value]]:
        """
        Returns a new parser where the empty string means None

        Args:
            strip: Whether to strip whitespace

        Returns:
            A new parser
        """
        return _EmptyMeansNone(self, strip)

    def separated_by(
        self, sep: str, strip: bool = True, remove_empty: bool = True
    ) -> Parser[Sequence[_Value]]:
        """
        Returns a new parser that parses values separated by a string

        Args:
            sep: Separator
            strip: Whether to strip whitespace from separated values
            remove_empty: Whether to remove empty strings before parsing them

        Returns:
            A new parser
        """
        return _SeparatedBy(self, sep, strip, remove_empty)

    def validated(
        self, predicate: Callable[[_Value], bool], msg: Union[str, Callable[[str, _Value], str]]
    ) -> _Validated[_Value]:
        """
        Returns a parser that verifies a predicate after successful parse

        Args:
            predicate: Predicate to check
            msg: Either a string, or a function that constructs a string from the parsed string
                 and the result

        Returns:
            A new parser
        """
        if isinstance(msg, str):
            c: str = msg
            assert isinstance(predicate, Predicate)
            return _Validated(self, predicate, lambda a, t: c)
        else:
            assert isinstance(predicate, Predicate)
            return _Validated(
                self,
                predicate,
                msg,
            )

    @staticmethod
    def from_parsy_parser(parser: parsy.Parser) -> Parser[_Value]:
        """
        Creates a parser from a parsy parser

        Args:
            parser: Parser returning a value of type ``t``

        Returns:
            Parameter type
        """
        import parsy  # pylint: disable=import-outside-toplevel,redefined-outer-name

        @dataclass(frozen=True)
        class _Parsy(Parser[_Value]):
            """
            Wraps a parser from the parsy library
            """

            parser: parsy.Parser

            def parse(self, arg: str) -> Res[_Value]:
                res = (self.parser << parsy.eof)(arg, 0)  # Inspired by Parser.parse
                if res.status:
                    return cast(_Value, res.value)
                else:
                    if res.furthest is not None:
                        return Err.make(
                            f"Parse error '{res.expected}' in '{arg}' at position '{res.furthest}'"
                        )
                    else:
                        return Err.make(f"Parse error '{res.expected}' in '{arg}'")

        return _Parsy(parser)

    @staticmethod
    def from_function_that_raises(
        f: Callable[[str], _Value], *catch: Type[Exception]
    ) -> Parser[_Value]:
        """
        Creates a parser from a function that raises exceptions on parse errors

        Args:
            f: Function that parses the string
            catch: Exception types to catch; if no types are provided, all exceptions will be
                   caught

        Returns:
            Parameter type
        """
        return _FromFunctionThatRaises(f, catch)

    @staticmethod
    def from_function(f: Callable[[str], Res[_Value]]) -> Parser[_Value]:
        """
        Creates a parser from a function that returns a value or an error

        Args:
            f: Function that parses the string

        Returns:
            Parameter type
        """

        return _FromFunction(f)

    @staticmethod
    def from_choices(
        values: Iterable[str],
        strip: bool = True,
        force_case: ForceCase = ForceCase.NO_CHANGE,
    ) -> Parser[str]:
        """
        Creates a parser whose values are chosen from a set of strings

        Args:
            values: Set of values
            strip: Whether to strip whitespace before looking for choices

        Returns:
            Parameter type
        """

        return Parser.from_mapping({v: v for v in values}, strip, force_case)

    @staticmethod
    def from_mapping(
        mapping: Mapping[str, _Value],
        strip: bool = True,
        force_case: ForceCase = ForceCase.NO_CHANGE,
        aliases: Optional[Mapping[str, _Value]] = None,
    ) -> Parser[_Value]:
        """
        Creates a parser whose strings correspond to keys in a dictionary

        Args:
            mapping: Dictionary mapping strings to values
            strip: Whether to strip whitespace before looking for keys
            force_case: Whether to normalize the case of the user string
            aliases: Additional mappings not shown in help

        Returns:
            Parameter type
        """
        if aliases is None:
            aliases = {}
        return _Choices(mapping, strip, force_case, aliases)


@dataclass(frozen=True)
class _Choices(Parser[_Value]):
    """
    Describes a multiple choice parser
    """

    mapping: Mapping[str, _Value]
    strip: bool
    force_case: ForceCase
    aliases: Mapping[str, _Value]

    def parse(self, arg: str) -> Res[_Value]:
        if self.strip:
            arg = arg.strip()
        if self.force_case is ForceCase.LOWER:
            arg = arg.lower()
        elif self.force_case is ForceCase.UPPER:
            arg = arg.upper()
        elif self.force_case is ForceCase.NO_CHANGE:
            pass
        else:
            assert_never(self.force_case)
        all_mappings = {**self.mapping, **self.aliases}
        if arg in all_mappings:
            return all_mappings[arg]
        else:
            msg = f"Value {arg} not in choices {','.join(self.mapping.keys())}"
            return Err.make(msg)

    def choices(self) -> Optional[Sequence[str]]:
        return list(self.mapping.keys())


@dataclass  # not frozen because mypy bug, please be responsible
class _FromFunctionThatRaises(Parser[_Value]):
    """
    Wraps a function that may raise exceptions
    """

    # the optional is to make mypy happy
    fun: Callable[[str], _Value]  #: Callable function that may raise
    catch: Sequence[Type[Exception]]

    def parse(self, arg: str) -> Res[_Value]:
        try:
            f = self.fun
            assert f is not None
            return f(arg)
        except Exception as e:  # pylint: disable=broad-except
            if (not self.catch) or any([isinstance(e, t) for t in self.catch]):
                return Err.make(f"Error '{e}' in '{arg}'")
            else:
                raise


@dataclass  # not frozen because mypy bug, please be responsible
class _FromFunction(Parser[_Value]):
    """
    Wraps a function that returns a result
    """

    fun: Callable[[str], Res[_Value]]

    def parse(self, arg: str) -> Res[_Value]:
        return self.fun(arg)


@dataclass(frozen=True)
class _EmptyMeansNone(Parser[Optional[_Item]]):
    """
    Wraps an existing parser so that "empty means none"
    """

    wrapped: Parser[_Item]  #: Wrapped Parser called if value is not empty
    strip: bool  #:  Whether to strip whitespace before testing for empty

    def parse(self, arg: str) -> Res[Optional[_Item]]:
        if self.strip:
            arg = arg.strip()
        if not arg:
            return None
        else:
            return self.wrapped.parse(arg)


@dataclass(frozen=True)
class _FlatMapped(Parser[_ReturnType], Generic[_Parameter, _ReturnType]):
    """
    Wraps an existing parser and applies a function to its successful result
    """

    wrapped: Parser[_Parameter]  #: Wrapped parser

    #: Mapping function, made optional as a hack
    f: Optional[Callable[[_Parameter], Res[_ReturnType]]]

    def parse(self, arg: str) -> Res[_ReturnType]:
        assert self.f is not None
        res: Res[_Parameter] = self.wrapped.parse(arg)
        if isinstance(res, Err):
            return res
        return self.f(res)

    def choices(self) -> Optional[Sequence[str]]:
        return self.wrapped.choices()


@dataclass(frozen=True)
class _Mapped(Parser[_ReturnType], Generic[_Parameter, _ReturnType]):
    """
    Wraps an existing parser and applies a function to its successful result
    """

    wrapped: Parser[_Parameter]  #: Wrapped parser
    f: Optional[Callable[[_Parameter], _ReturnType]]  #: Mapping function, made optional as a hack

    def parse(self, arg: str) -> Res[_ReturnType]:
        assert self.f is not None
        res: Res[_Parameter] = self.wrapped.parse(arg)
        if isinstance(res, Err):
            return res
        return self.f(res)

    def choices(self) -> Optional[Sequence[str]]:
        return self.wrapped.choices()


@dataclass(frozen=True)
class _SeparatedBy(Parser[Sequence[_Item]]):
    """
    Parses values separated by a given separator
    """

    item: Parser[_Item]  #: Parser of individual items
    sep: str  #: Item separator
    strip: bool  #: Whether to strip whitespace around separated strings
    remove_empty: bool  #: Whether to prune empty strings

    def parse(self, arg: str) -> Res[Sequence[_Item]]:
        items: Iterable[str] = arg.split(self.sep)
        if self.strip:
            items = map(lambda s: s.strip(), items)
        if self.remove_empty:
            items = filter(None, items)
        res: Sequence[Res[_Item]] = [self.item.parse(s) for s in items]

        return collect_seq(res)


@dataclass(frozen=True)
class _Validated(Parser[_Value]):
    """
    Parses a value and validates it
    """

    wrapped: Parser[_Value]
    predicate: Predicate[_Value]
    msg: ErrorMessageProvider[_Value]

    def parse(self, arg: str) -> Res[_Value]:
        res = self.wrapped.parse(arg)
        if isinstance(res, Err):
            return res
        p = self.predicate
        if p(res):
            return res
        return Err.make(self.msg(arg, res))

    def choices(self) -> Optional[Sequence[str]]:
        return self.wrapped.choices()


#: Parses a path
path_parser: Parser[pathlib.Path] = Parser.from_function_that_raises(pathlib.Path)

#: Parses an integer
int_parser: Parser[int] = Parser.from_function_that_raises(int)

#: Parses a string, preserves whitespace
str_parser: Parser[str] = Parser.from_function(lambda s: s)

#: Parses a string, stripping whitespace left and right
stripped_str_parser: Parser[str] = Parser.from_function(lambda s: s.strip())

#: Parses a float
float_parser: Parser[float] = Parser.from_function_that_raises(float)

bool_parser: Parser[bool] = Parser.from_mapping(
    {"true": True, "false": False},
    force_case=ForceCase.LOWER,
    aliases={"t": True, "f": False, "1": True, "0": False},
)



"""
Error-reporting module

This implements a union-based result type similar to `<https://github.com/rustedpy/result/>`_,
except that we do not wrap the "good" result.

The type :data:`.Res` contains either a value of a user-defined type :data:`._Value_co`, an error
or a list of errors.

Errors have an associated context, see the description for :class:`.Err1`.

.. rubric:: Types

.. py:data:: Res

    Result type (parameterized :data:`.Res` [ :data:`._Value_co` ] )

    Note:
        Wrapping an error in :data:`.Res` as in :data:`.Res` [ :class:`.Err` ] is meaningless,
        and will not do what you expect.

        There are other "monadic" result types for Python that have better composability.

    :data:`.Res` = :data:`~typing.Union` [ :data:`._Value_co`, :class:`Err` ]

.. py:data:: _Value_co

    OK Value in our custom result type

.. py:data:: _Parameters

    Parameter specification for a decorated function

.. py:data:: _ReturnType

    Return type for a decorated function

.. py:data:: _U

    Generic type

.. py:data:: _V

    Generic type

.. py:data:: _W

    Generic type

"""

from __future__ import annotations

import shutil
import textwrap
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import ParamSpec


class Err(ABC):
    """
    Describes either an error or a list of errors
    """

    @abstractmethod
    def markdown(self) -> Sequence[str]:
        """
        Returns a Markdown-formatted summary of this error (or list of errors)
        """

    @abstractmethod
    def errors(self) -> Sequence[Err1]:
        """
        Returns a sequence of all contained errors

        If this is not a collection of errors :class:`.ManyErr`, returns a sequence with
        a single item, this instance itself.
        """

    @abstractmethod
    def in_context(self, **contexts: Any) -> Err:
        """
        Adds to this error information about the context in which it occurred

        Args:
            contexts: Contexts (given as key/value pairs) to append to the context list

        Returns:
            Updated error
        """

    def pretty_print(self) -> None:
        """
        Pretty prints an error on the console
        """
        try:
            from rich.console import Console  # pylint: disable=import-outside-toplevel
            from rich.markdown import Markdown  # pylint: disable=import-outside-toplevel

            console = Console()
            md = Markdown("\n".join(self.markdown()))
            console.print(md)
        except ImportError:
            sz = shutil.get_terminal_size()
            t = self.markdown()
            print(textwrap.fill("\n".join(t), width=sz.columns))

    @staticmethod
    def collect1(first_error: Err, *additional_errors: Err) -> Err:
        """
        Collect a non-empty sequence of errors into a single error

        Example:
            We can also collect errors coming from a list using the following syntax

            >>> errors = [Err.make('err 1'), Err.make('err 2'), Err.make('err 3')]
            >>> Err.collect1(*errors)
            ManyErr(errs=...)

        Args:
            first_error: First error to collect
            additional_errors: Errors to collect into a single error, at least one must be provided

        Raises:
            ValueError: If no error is provided

        Returns:
            A consolidated error
        """
        res = Err.collect(first_error, *additional_errors)
        if res is None:
            raise ValueError("Result cannot be None if at least one error is provided")
        return res

    @staticmethod
    def collect(*errs: Optional[Err]) -> Optional[Err]:
        """
        Collect a possibly empty sequence of (optional) errors into an optional single error

        Args:
            errs: (Optional) errors

        Returns:
            An error or `None`
        """

        lst: List[Err1] = []
        for e in errs:
            if e is not None:
                lst.extend(e.errors())
        if not lst:
            return None
        elif len(lst) == 1:
            return lst[0]
        else:
            return ManyErr(lst)

    @staticmethod
    def make(msg: str, **contexts: Any) -> Err1:
        """
        Creates a single error

        Example:
            >>> Err.make("test error")
            Err1(msg='test error', contexts=[])

        Args:
            msg: Error message
            contexts: Contexts (given as key/value pairs) to append to the context list

        Returns:
            An error
        """
        return Err1(msg, [*contexts.items()])

    @staticmethod
    def check(predicate: bool, msg: str, **contexts: Any) -> Optional[Err1]:
        """
        Returns an error if the given predicate is false, otherwise returns None.

        Example:
            >>> a = -1
            >>> b = 2
            >>> Err.collect(Err.check(a >= 0, "a < 0"), Err.check(b >= 0, "b < 0"))
            Err1(msg='a < 0', contexts=[])

        Args:
            predicate: Predicate to check
            msg: Error message
            contexts: Contexts (given as key/value pairs) to append to the context list

        Returns:
            An error or `None`
        """
        if predicate:
            return None
        else:
            return Err.make(msg, **contexts)


@dataclass(frozen=True)
class Err1(Err):
    """
    Describes a single error

    This error stores a sequence of contexts, which are key/value pairs describing in which
    context the error occurred. Contexts are used for explaining to the user where the error
    occurred. They shouldn't be used, in principle, to wrap additional information about the error.
    """

    msg: str  #: Error message
    contexts: Sequence[Tuple[str, Any]]  #: Contexts in which the error appears, from old to new

    def errors(self) -> Sequence[Err1]:
        return [self]

    def markdown(self, include_contexts: bool = True) -> Sequence[str]:
        c: List[str] = []
        if include_contexts:
            c = [
                line
                for (name, value) in reversed(self.contexts)
                for line in [f"In {name}: {value}", ""]
            ]
        return [*c, "`" + self.msg + "`"]

    def in_context(self, **contexts: Any) -> Err1:
        return Err1(self.msg, [*contexts.items(), *self.contexts])


@dataclass(frozen=True)
class _GroupedErrors:
    """
    Used to group errors by context when pretty-printing

    Forms a tree structure: at a level of the tree, errors which share a similar "top" context
    are grouped; other errors are kept in an ungrouped list.

    At the next level, the top-level context is popped, and the process repeat.
    """

    groups: Mapping[Tuple[str, Any], _GroupedErrors]  #: Errors grouped by context
    ungrouped: Sequence[Err1]  #: Errors not grouped by context

    def markdown(
        self,
        skip_contexts: int = 0,
    ) -> Sequence[str]:
        """
        Returns the markdown formatted info about this group of errors

        Args:
            skip_contexts: How many context to skip because they are displayed by the surrounding
                           text
        """
        lines: List[str] = []
        for ctx, ge in self.groups.items():
            ctx_name, ctx_value = ctx
            lines.append(f"* In {ctx_name}: {ctx_value}")
            lines.append("")
            lines.extend(["  " + l for l in ge.markdown(skip_contexts + 1)])
        d = len(str(len(self.ungrouped) - 1)) + 1
        for i, e in enumerate(self.ungrouped):
            err_lines: List[str] = []
            ctxs = list(e.contexts)[skip_contexts:]
            for ctx in ctxs:
                ctx_name, ctx_value = ctx
                err_lines.append(f"In {ctx_name}: {ctx_value}")
                err_lines.append("")
            err_lines.extend(e.markdown(include_contexts=False))
            num = str(i) + "." + " " * (d - len(str(i)))
            lines.append(num + err_lines[0])
            sp = " " * (d + 1)
            lines.extend([sp + l for l in err_lines[1:]])
            lines.append("")
        return lines

    @staticmethod
    def make(errors: Sequence[Err1], index: int = 0) -> _GroupedErrors:
        """
        Constructs a tree of grouped errors

        Args:
            errors: Sequence of errors
            index: Number of contexts to skip, which have already been processed
        """
        grouped_pairs: Dict[Tuple[str, Any], List[Tuple[int, Err1]]] = {}
        ungrouped_pairs: List[Tuple[int, Err1]] = []
        for i, e in enumerate(errors):
            if index < len(e.contexts):
                ctx = e.contexts[index]
                if isinstance(ctx, Hashable):
                    if ctx not in grouped_pairs:
                        grouped_pairs[ctx] = [(i, e)]
                    else:
                        grouped_pairs[ctx].append((i, e))
            else:
                ungrouped_pairs.append((i, e))
        groups: Dict[Tuple[str, Any], _GroupedErrors] = {}
        for k, v in grouped_pairs.items():
            if len(v) == 1:
                ungrouped_pairs.append((v[0]))
            else:
                groups[k] = _GroupedErrors.make([e for i, e in v], index + 1)

        def get_index(pair: Tuple[int, Err1]) -> int:
            return pair[0]

        ungrouped: List[Err1] = [e for i, e in sorted(ungrouped_pairs, key=get_index)]

        return _GroupedErrors(groups, ungrouped)


@dataclass(frozen=True)
class ManyErr(Err):
    """
    Error class that regroups several errors

    It must contain at least one error, and it makes sense to use it when there is more than one
    error present.

    This class has special support for pretty-printing a list of errors.
    """

    errs: Sequence[Err1]  #: Contained errors

    def __post_init__(self) -> None:
        assert len(self.errs) > 0, "A ManyErr should contain at least one error"
        assert all([not isinstance(e, ManyErr) for e in self.errs])

    def markdown(self) -> Sequence[str]:
        return _GroupedErrors.make(self.errs).markdown()

    def errors(self) -> Sequence[Err1]:
        return self.errs

    def in_context(self, **contexts: Any) -> Err:
        return ManyErr([e.in_context(**contexts) for e in self.errs])


# those attributes are documented in the module docstring

_U = TypeVar("_U")

_V = TypeVar("_V")

_W = TypeVar("_W")

_Value_co = TypeVar("_Value_co", covariant=True)

_Parameters = ParamSpec("_Parameters")

_ReturnType = TypeVar("_ReturnType")

Res = Union[_Value_co, Err]


def wrap(
    *keep: Type[Exception],
) -> Callable[[Callable[_Parameters, _ReturnType]], Callable[_Parameters, Res[_ReturnType]],]:
    """
    Decorates the given function, and wraps thrown exceptions into an error result

    Exception types can be provided as arguments to filter the exceptions being transformed
    into errors.

    Example:
        >>> @wrap(ValueError)
        ... def parse(s: str) -> int:
        ...     return int(s)
        >>> parse("2")
        2
        >>> parse("should error")
        Err1(msg="...", contexts=[])
    """

    def decorator(
        func: Callable[_Parameters, _ReturnType]
    ) -> Callable[_Parameters, Res[_ReturnType]]:
        def wrapper(*args, **kwargs) -> Res[_ReturnType]:  # type: ignore
            try:
                res: Res[_ReturnType] = func(*args, **kwargs)
            except Exception as e:  # pylint: disable=broad-except
                if (not keep) or any([isinstance(e, t) for t in keep]):
                    res = Err.make(f"{type(e).__name__} thrown: {e}")
                else:
                    raise
            return res

        return wrapper

    return decorator


@overload
def in_context(result: Optional[Err], **contexts: Any) -> Optional[Err]:
    pass


@overload
def in_context(result: Res[_Value_co], **contexts: Any) -> Res[_Value_co]:
    pass


def in_context(
    result: Union[_Value_co, Err, None], **contexts: Any
) -> Union[_Value_co, Err, None]:
    """
    Adds context to an error contained in a result type (when applicable)

    This function has different signatures depending on the argument types.

    Args:
        result: Result to enrich, if it contains an error
        contexts: Contexts to add, given as keyword arguments

    :type result: See signatures

    Returns:
        Updated result with added context

    :rtype: See signatures

    See also:
        :meth:`.Err.in_context`
    """
    if isinstance(result, Err):
        return result.in_context(**contexts)
    else:
        return result


@overload
def collect(t: Res[_Value_co], u: Res[_U]) -> Res[Tuple[_Value_co, _U]]:
    pass


@overload
def collect(t: Res[_Value_co], u: Res[_U], v: Res[_V]) -> Res[Tuple[_Value_co, _U, _V]]:
    pass


@overload
def collect(
    t: Res[_Value_co], u: Res[_U], v: Res[_V], w: Res[_W]
) -> Res[Tuple[_Value_co, _U, _V, _W]]:
    pass


def collect(*args):  # type: ignore[no-untyped-def]
    """
    Collects single results of various types into a tuple result


    :rtype: See signatures
    """
    ok: List[Any] = []
    errs: List[Err1] = []
    for arg in args:
        if isinstance(arg, Err):
            errs.extend(arg.errors())
        else:
            ok.append(arg)
    if errs:
        return Err.collect1(*errs)  # pylint: disable=no-value-for-parameter
    else:
        return tuple(ok)


def map(  # pylint: disable=redefined-builtin
    f: Callable[[_Value_co], _ReturnType], r: Res[_Value_co]
) -> Res[_ReturnType]:
    """
    Enables a computation on the value of a result

    Example:

        >>> from math import sqrt
        >>> @wrap(ValueError)
        ... def parse(s: str) -> float:
        ...    return float(s)
        >>> @wrap(ValueError)
        ... def square(f: float) -> float:
        ...    return f*f
        >>> map(square, parse("2"))
        4.0
        >>> map(square, parse("-1"))
        1.0
        >>> flat_map(square, parse("invalid"))
        Err1(msg="ValueError thrown: could not convert string to float: 'invalid'", contexts=[])

    Args:
        f: Function that takes a value of type `._Value_co` and returns a value of type
           `._ReturnType`
        r: Result parameterized by type `._Value_co`

    Returns:
        Result parameterized by type `._ReturnType`

    See also:
        :func:`.flatMap`
    """
    if isinstance(r, Err):
        return r
    else:
        return f(r)


def flat_map(f: Callable[[_Value_co], Res[_ReturnType]], r: Res[_Value_co]) -> Res[_ReturnType]:
    """
    Enables the chaining computations that can error

    Example:

        >>> from math import sqrt
        >>> @wrap(ValueError)
        ... def parse(s: str) -> float:
        ...    return float(s)
        >>> @wrap(ValueError)
        ... def my_sqrt(f: float) -> float:
        ...    return sqrt(f)
        >>> flat_map(my_sqrt, parse("2"))
        1.414...
        >>> flat_map(my_sqrt, parse("-1"))
        Err1(msg='ValueError thrown: math domain error', contexts=[])
        >>> flat_map(my_sqrt, parse("invalid"))
        Err1(msg="ValueError thrown: could not convert string to float: 'invalid'", contexts=[])

    Args:
        f: Function that takes a result parameterized by type `_Value` and returns a result
           parameterized by type `_ReturnType`
        r: Result parameterized by type `_Value`

    Returns:
        Result parameterized by type `_ReturnType`

    See also:
        :func:`.map`
    """
    if isinstance(r, Err):
        return r
    else:
        return f(r)


def flatMap(f: Callable[[_Value_co], Res[_ReturnType]], r: Res[_Value_co]) -> Res[_ReturnType]:
    """
    Enables the chaining computations that can error (deprecated)

    See also: :func:`.flat_map`
    """
    warnings.warn(
        "flatMap has been deprecated, use flat_map instead", DeprecationWarning, stacklevel=2
    )
    return flat_map(f, r)


def collect_seq(seq: Sequence[Res[_Value_co]]) -> Res[Sequence[_Value_co]]:
    """
    Collects a sequence of results

    If any of the sequence elements is an error, the result is an error which collects all
    the errors present in the sequence.

    If all the sequence elements are values, it returns a sequence of values.

    Example:

        >>> @wrap(ValueError)
        ... def parse(s: str) -> int:
        ...    return int(s)
        >>> collect_seq([parse("2"), parse("invalid")])
        Err1(...)
        >>> collect_seq([parse("2"), parse("3")])
        [2, 3]


    Args:
        seq: Sequence of results

    Returns:
        An error or a sequence
    """
    ok: List[_Value_co] = []
    errs: List[Err1] = []
    for res in seq:
        if isinstance(res, Err):
            errs.extend(res.errors())
        else:
            ok.append(res)
    if errs:
        return Err.collect1(*errs)  # pylint: disable=unused-import
    else:
        return ok



"""
Argument definition

This module enables the definitions of various kinds of configuration arguments.

- Parameters (:class:`.Param`) correspond to values that will be present in the configuration.
  Each invocation of a parameter has a user-provided string attached, which is parsed by a
  :class:`~configpile.parsers.Parser`. Those invocations can come from environment variables
  (:attr:`.Param.env_var_name`), configuration files (:attr:`.Param.config_key_name`),
  command-line flags (:attr:`.Arg.short_flag_name`, :attr:`.Arg.long_flag_name`), or finally
  as positional command-line parameters (:attr:`.Param.positional`).

- Configuration file inclusion using special parameters :meth:`.Param.config`

- Expanders (:class:`.Expander`) can only be present as command-line flags; their action is
  to insert a key/value pair in the command-line. This enables, for example, flags that set
  a boolean parameter to `True` or `False`.

.. rubric:: Types

This module uses the following types.

.. py:data:: _Config

    Configuration dataclass type being constructed

.. py:data:: _Arg

    Precise argument type, used mostly internally.

    See `<https://en.wikipedia.org/wiki/Bounded_quantification>`_

.. py:data:: _Value_co

    Type of the parameter value

.. py:data:: _Item_co

    Item in a sequence
"""

from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from .collector import Collector
from .enums import Derived, Positional
from .handlers import CLConfigParam, CLInserter, CLParam, KVConfigParam, KVParam
from .parsers import Parser, path_parser

# documented in the module docstring

_Value_co = TypeVar("_Value_co", covariant=True)

_Item_co = TypeVar("_Item_co", covariant=True)

if TYPE_CHECKING:
    from .config import Config
    from .processor import ProcessorFactory


_Arg = TypeVar("_Arg", bound="Arg")
_Config = TypeVar("_Config", bound="Config")


@dataclass(frozen=True)
class Arg:
    """
    Base class for all kinds of arguments
    """

    #: Help for the argument
    help: Optional[str]

    #: Short option name, used in command line parsing, prefixed by a single hyphen
    short_flag_name: Optional[str]

    #: Long option name used in command line argument parsing
    #:
    #: It is in general lowercase, prefixed with ``--`` and words are separated by hyphens.,
    #:
    #: If set to
    long_flag_name: Union[str, Derived, None]

    def __post_init__(self) -> None:
        if self.__class__ == Arg:
            raise NotImplementedError("Cannot instantiate Arg directly")

    def all_flags(self) -> Sequence[str]:
        """
        Returns a sequence of all forms of command line flags
        """
        res: List[str] = []
        if self.short_flag_name is not None:
            res.append(self.short_flag_name)
        assert not isinstance(self.long_flag_name, Derived)
        if isinstance(self.long_flag_name, str):
            res.append(self.long_flag_name)
        return res

    def update_dict_(
        self,
        name: str,
        help: Optional[str],  # pylint: disable=redefined-builtin
        env_prefix: Optional[str],
    ) -> Mapping[str, Any]:
        """
        Returns updated values for this argument, used during :class:`.App` construction

        Args:
            name: Argument field name
            help: Argument docstring which describes the argument role
            env_prefix: Uppercase prefix for all environment variables

        Returns:

        """
        res: Dict[str, Any] = {}
        if self.help is None:
            res = {"help": help}
        if isinstance(self.long_flag_name, Derived):
            res["long_flag_name"] = "--" + self.long_flag_name.derive(name)
        return res

    def updated(
        self: _Arg,
        name: str,
        help: str,  # pylint: disable=redefined-builtin
        env_prefix: Optional[str],
    ) -> _Arg:
        """
        Returns a copy of this argument with some data updated from its declaration context

        Args:
            self:
            name: Identifier name
            help: Help string (derived from autodoc syntax)
            env_prefix: Environment prefix

        Returns:
            Updated argument
        """
        return dataclasses.replace(self, **self.update_dict_(name, help, env_prefix))

    def update_processor(self, pf: ProcessorFactory[_Config]) -> None:
        """
        Updates a config processor with the processing required by this argument

        Args:
            pf: Processor factory
        """
        raise NotImplementedError

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        """
        Returns the keyword arguments for use with argparse.ArgumentParser.add_argument

        Returns:
            Keyword arguments mapping
        """
        raise NotImplementedError


@dataclass(frozen=True)
class Expander(Arg):
    """
    Command-line argument that expands into a flag/value pair
    """

    new_flag: str  #: Inserted flag in the command line

    new_value: str  #: Inserted value in the command line

    def inserts(self) -> Tuple[str, str]:
        """
        Returns the flag/value pair that is inserted when this command flag is present
        """
        return (self.new_flag, self.new_value)

    def update_processor(self, pf: ProcessorFactory[_Config]) -> None:
        for flag in self.all_flags():
            pf.cl_flag_handlers[flag] = CLInserter([self.new_flag, self.new_value])
        pf.ap_commands.add_argument(*self.all_flags(), **self.argparse_argument_kwargs())

    @staticmethod
    def make(
        new_flag: str,
        new_value: str,
        *,
        help: Optional[str] = None,  # pylint: disable=redefined-builtin
        short_flag_name: Optional[str] = None,
        long_flag_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
    ) -> Expander:
        """
        Constructs an expander that inserts a flag/value pair in the command line

        At least one of ``short_flag_name`` or ``long_flag_name`` must be defined.

        Args:
            new_flag: Inserted flag, including the hyphen prefix
            new_value: String value to insert following the flag

        Keyword Args:
            help: Help description (autodoc/docstring is used otherwise)
            short_flag_name: Short flag name of this command flag
            long_flag_name: Long flag name of this command flag
        """
        res = Expander(
            help=help,
            new_flag=new_flag,
            new_value=new_value,
            short_flag_name=short_flag_name,
            long_flag_name=long_flag_name,
        )
        return res

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        return {"help": self.help}


@dataclass(frozen=True)
class Param(Arg, Generic[_Value_co]):
    """
    Describes an argument holding a value of a given type

    You'll want to construct parameters using the static methods available:

    - :meth:`~.Param.store` to create a parameter that keeps the last value provided

    - :meth:`~.Param.append` to create a parameter that collects all the provided values into
      a sequence. Note that the parsed value must already be a sequence.

    - :meth:`~.Param.append1` to create a parameter that collects all the provided value into
      a sequence. The parsed value should not be a sequence.

    For the constructions above, you can construct positional parameters by setting the
    :attr:`.Param.positional` field.

    - :meth:`~.Param.config` returns a parameter that parses configuration files.
    """

    #: Argument type, parser from string to value
    parser: Parser[_Value_co]  # type: ignore

    is_config: bool  #: Whether this represent a list of config files

    #: Argument collector
    collector: Collector[_Value_co]  # type: ignore

    default_value: Optional[str]  #: Default value inserted as instance

    #: Python field identifier for this parameter in :class:`~configpile.config.Config`
    name: Optional[str]  #: Python identifier representing the argument

    #: Whether this parameter can appear as a positional parameter and how
    #:
    #: A positional parameter is a parameter that appears without a preceding flag
    positional: Optional[Positional]

    #: Configuration key name used in INI files
    #:
    #: It is lowercase, and words are separated by hyphens.
    config_key_name: Union[str, Derived, None]

    #: Environment variable name
    #:
    #: The environment variable name has a prefix, followed by the
    #: Python identifier in uppercase, with underscore as separator.
    #:
    #: This prefix is provided by :attr:`.App.env_prefix_`
    env_var_name: Union[str, Derived, None]

    def update_dict_(
        self,
        name: str,
        help: Optional[str],  # pylint: disable=redefined-builtin
        env_prefix: Optional[str],
    ) -> Mapping[str, Any]:
        r = {"name": name, **super().update_dict_(name, help, env_prefix)}
        if isinstance(self.config_key_name, Derived):
            r["config_key_name"] = self.config_key_name.derive(name)
        if isinstance(self.env_var_name, Derived) and env_prefix is not None:
            r["env_var_name"] = env_prefix + self.env_var_name.derive(name)
        return r

    def all_config_key_names(self) -> Sequence[str]:
        """
        Returns a sequence of all forms of command line options

        Returns:
            Command line options
        """
        if isinstance(self.config_key_name, str):
            return [self.config_key_name]
        else:
            return []

    def all_env_var_names(self) -> Sequence[str]:
        """
        Returns a sequence of all forms of command line options

        Returns:
            Command line options
        """
        if isinstance(self.env_var_name, str):
            return [self.env_var_name]
        else:
            return []

    def is_required(self) -> bool:
        """
        Returns whether the argument is required
        """
        return self.default_value is None and self.collector.arg_required()

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        res: Dict[str, Any] = {"help": self.help}
        choices = self.parser.choices()
        if choices is not None:
            res = {**res, "choices": choices, "type": str}
        return {
            **res,
            **self.collector.argparse_argument_kwargs(),
        }

    def update_processor(self, pf: ProcessorFactory[_Config]) -> None:

        assert self.name is not None
        pf.params_by_name[self.name] = self
        if self.positional is not None:
            pf.cl_positionals.append(self)
        for flag in self.all_flags():
            if self.is_config:
                pf.cl_flag_handlers[flag] = CLConfigParam(self)
            else:
                pf.cl_flag_handlers[flag] = CLParam(self)

        for key in self.all_config_key_names():
            pf.ini_handlers[key] = KVParam(self)
        for name in self.all_env_var_names():
            if self.is_config:
                pf.env_handlers[name] = KVConfigParam(self)
            else:
                pf.env_handlers[name] = KVParam(self)

        flags = self.all_flags()
        if self.is_required():
            pf.ap_required.add_argument(*flags, dest=self.name, **self.argparse_argument_kwargs())
        else:
            pf.ap_optional.add_argument(*flags, dest=self.name, **self.argparse_argument_kwargs())

    @staticmethod
    def store(
        parser: Parser[_Value_co],
        *,
        help: Optional[str] = None,  # pylint: disable=redefined-builtin
        default_value: Optional[str] = None,
        positional: Optional[Positional] = None,
        short_flag_name: Optional[str] = None,
        long_flag_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
        config_key_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
        env_var_name: Union[str, Derived, None] = None,
    ) -> Param[_Value_co]:
        """
        Creates a parameter that stores the last provided value

        If a default value is provided, the argument can be omitted. However,
        if the default_value ``None`` is given (default), then
        the parameter cannot be omitted.

        Args:
            parser: Parser that transforms a string into a value

        Keyword Args:
            help: Help description (autodoc/docstring is used otherwise)
            default_value: Default value
            positional: Whether this parameter is present in positional arguments
            short_flag_name: Short option name (optional)
            long_flag_name: Long option name (auto. derived from fieldname by default)
            config_key_name: Config key name (auto. derived from fieldname by default)
            env_var_name: Environment variable name (forbidden by default)

        Returns:
            The constructed Param instance
        """

        return Param(
            name=None,
            help=help,
            parser=parser,
            collector=Collector.keep_last(),
            default_value=default_value,
            positional=positional,
            short_flag_name=short_flag_name,
            long_flag_name=long_flag_name,
            config_key_name=config_key_name,
            env_var_name=env_var_name,
            is_config=False,
        )

    @staticmethod
    def config(
        *,
        help: Optional[str] = None,  # pylint: disable=redefined-builtin
        short_flag_name: Optional[str] = None,
        long_flag_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
        env_var_name: Union[str, Derived, None] = None,
    ) -> Param[Sequence[Path]]:
        """
        Creates a parameter that parses configuration files and stores their names

        Keyword Args:
            help: Help description (autodoc/docstring is used otherwise)
            short_flag_name: Short option name (optional)
            long_flag_name: Long option name (auto. derived from fieldname by default)
            env_var_name: Environment variable name (forbidden by default)

        Returns:
            A configuration files parameter
        """
        return Param(
            name=None,
            help=help,
            parser=path_parser.separated_by(",", strip=True, remove_empty=True),
            collector=Collector.append(),  # type: ignore
            positional=None,
            short_flag_name=short_flag_name,
            long_flag_name=long_flag_name,
            config_key_name=None,
            env_var_name=env_var_name,
            is_config=True,
            default_value=None,
        )

    @staticmethod
    def append1(
        parser: Parser[_Item_co],
        *,
        help: Optional[str] = None,  # pylint: disable=redefined-builtin
        positional: Optional[Positional] = None,
        short_flag_name: Optional[str] = None,
        long_flag_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
        config_key_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
        env_var_name: Union[str, Derived, None] = None,
    ) -> Param[Sequence[_Item_co]]:
        """
        Returns a newly created argument that appends single parsed values

        While :meth:`~Param.append` creates a parameter that joins sequences of values, this method
        creates a parameter that appends single items to a sequence.

        Args:
            parser: Parser that transforms a string into a single value item

        Keyword Args:
            help: Help description (autodoc/docstring is used otherwise)
            positional: Whether this argument is present in positional arguments
            short_flag_name: Short option name (optional)
            long_flag_name: Long option name (auto. derived from fieldname by default)
            config_key_name: Config key name (auto. derived from fieldname by default)
            env_var_name: Environment variable name (forbidden by default)
        """
        return Param(
            name=None,
            help=help,
            parser=parser.as_sequence_of_one(),
            collector=Collector.append(),  # type: ignore
            default_value=None,
            positional=positional,
            short_flag_name=short_flag_name,
            long_flag_name=long_flag_name,
            config_key_name=config_key_name,
            env_var_name=env_var_name,
            is_config=False,
        )

    @staticmethod
    def append(
        parser: Parser[Sequence[_Item_co]],
        *,
        help: Optional[str] = None,  # pylint: disable=redefined-builtin
        positional: Optional[Positional] = None,
        short_flag_name: Optional[str] = None,
        long_flag_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
        config_key_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
        env_var_name: Union[str, Derived, None] = None,
    ) -> Param[Sequence[_Item_co]]:
        """
        Returns a newly created argument that joins parsed sequences of values

        Args:
            parser: Parser that transforms a string into a sequence of values

        Keyword Args:
            help: Help description (autodoc/docstring is used otherwise)
            positional: Whether this argument is present in positional arguments
            short_flag_name: Short option name (optional)
            long_flag_name: Long option name (auto. derived from fieldname by default)
            config_key_name: Config key name (auto. derived from fieldname by default)
            env_var_name: Environment variable name (forbidden by default)
        """
        return Param(
            name=None,
            help=help,
            parser=parser,
            collector=Collector.append(),  # type: ignore
            default_value=None,
            positional=positional,
            short_flag_name=short_flag_name,
            long_flag_name=long_flag_name,
            config_key_name=config_key_name,
            env_var_name=env_var_name,
            is_config=False,
        )



"""
Example code of ``calculator6.py`` integrated into configpile as workaround
"""
import argparse
from dataclasses import dataclass
from math import prod
from pathlib import Path
from typing import Sequence

from typing_extensions import Annotated

from configpile import Config, Param, Positional, parsers


@dataclass(frozen=True)
class Calc(Config):
    """
    Command-line tool that sums an arbitrary number of floating point values
    """

    prog_ = "calculator6.py"

    config: Annotated[Sequence[Path], Param.config(help="INI configuration files to parse")]

    values: Annotated[
        Sequence[float],
        Param.append1(
            parsers.float_parser,
            positional=Positional.ZERO_OR_MORE,
            short_flag_name=None,
            long_flag_name=None,
            help="Values to sum",
        ),
    ]

    operation: Annotated[
        str,
        Param.store(
            parsers.Parser.from_choices(["+", "*"]),
            default_value="+",
            help='Operation to perform, either addition "+" or multiplication "*"',
        ),
    ]

    digits: Annotated[
        int,
        Param.store(
            parsers.int_parser.validated(lambda i: i > 0, "Must be a positive integer"),
            default_value="3",
            env_var_name="DIGITS",
            help="""
Number of digits to display

This number of digits can also be set with the environment variable
``DIGITS``
            """,
        ),
    ]


def argument_parser() -> argparse.ArgumentParser:
    """
    Returns a parser for use for sphinx-argparse
    """
    return Calc.get_argument_parser_()


if __name__ == "__main__":
    c = Calc.from_command_line_()
    fmt_string = f"%.{c.digits}f"
    if c.operation == "+":
        print(fmt_string % sum(c.values))
    elif c.operation == "*":
        print(fmt_string % prod(c.values))



"""
Configuration definition

This module defines the :class:`.Config` base class, from which all configuration derive.

.. rubric:: Types

This module uses the following types.

.. py:data:: _Config

    Configuration dataclass type being constructed

"""

from __future__ import annotations

import argparse
import inspect
import os
import sys
import typing
from abc import ABC
from dataclasses import dataclass
from pathlib import Path

from .processor import Processor, SpecialAction
from .userr import Err, Res

_Config = typing.TypeVar("_Config", bound="Config")


@dataclass(frozen=True)
class IniSection:
    """
    Describes a section of an INI file to include in the current configuration
    """

    name: str  #: Section name
    strict: bool  #: Whether all the keys must correspond to parsed arguments


@dataclass(frozen=True)
class Config(ABC):
    """
    Base class for dataclasses holding configuration data

    Example:
        .. code-block:: python

            @dataclass(frozen=True)
            class Adder(Config):
                x: Annotated[int, Param.store(types.int_)]
                y: Annotated[int, Param.store(types.int_, default_value="0")]

            res = Adder.from_command_line_(args=['x','2', 'y','3'], env={})
            res.x + res.y
    """

    # region Config: INI section processing

    #: Names of sections to parse in configuration files, with unknown keys ignored
    ini_relaxed_sections_: typing.ClassVar[typing.Sequence[str]] = ["Common", "COMMON", "common"]

    #: Names of additional sections to parse in configuration files, unknown keys error
    ini_strict_sections_: typing.ClassVar[typing.Sequence[str]] = []

    @classmethod
    def ini_sections_(cls) -> typing.Sequence[IniSection]:
        """
        Returns a sequence of INI file sections to parse

        By default, this parses first the relaxed sections and then the strict ones.

        First, try to replace the contents of :attr:`.ini_relaxed_sections_`
        and :attr:`.ini_strict_sections_`. Otherwise, override this method.
        """
        relaxed = [IniSection(name, False) for name in cls.ini_relaxed_sections_]
        strict = [IniSection(name, True) for name in cls.ini_strict_sections_]
        return relaxed + strict

    # endregion

    # region Config: general information about the program being configured

    prog_: typing.ClassVar[typing.Optional[str]] = None  #: Program name

    #: Text to display before the argument help
    #:
    #: If not present, taken from the Config subclass docstring.
    description_: typing.ClassVar[typing.Optional[str]] = None

    @classmethod
    def version_(cls) -> typing.Optional[str]:
        """
        Returns the version number of this script

        Designed to be overridden by a subclass

        Returns:
            Version number as a string
        """
        return None

    # endregion

    # region Config: environment variable processing

    #: Prefix for automatically derived environment variable names
    env_prefix_: typing.ClassVar[str] = ""

    # endregion

    # region Config: information constructed by configpile

    @classmethod
    def validators_(cls: typing.Type[_Config]) -> typing.Sequence[typing.Callable[[_Config], typing.Optional[Err]]]:
        """
        Returns all validators present in the given subclass of this class

        Validators are methods that take no arguments (except self) and return an optional error.
        Their name starts with ``validate_``.

        Returns:
            A sequence of all validators
        """
        res: typing.List[typing.Callable[[_Config], typing.Optional[Err]]] = []
        for name, _ in inspect.getmembers(cls, inspect.isroutine):
            if name.startswith("validate_"):
                res.append(getattr(cls, name))
        return res

    @classmethod
    def processor_(cls: typing.Type[_Config]) -> Processor[_Config]:
        """
        Returns a processor for this configuration
        """
        return Processor.make(cls)

    # endregion

    # region Config: configuration construction

    @classmethod
    def parse_ini_contents_(cls: typing.Type[_Config], ini_contents: str) -> Res[_Config]:
        """
        Parses the contents of an INI file into a configuration

        This method skips the processing of environment variables and command-line parameters.

        Args:
            ini_contents: Multiline string describing the configuration contents

        Returns:
            A parsed configuration or an error
        """
        processor = cls.processor_()
        return processor.process_ini_contents(ini_contents)

    @classmethod
    def parse_ini_file_(cls: typing.Type[_Config], ini_file_path: Path) -> Res[_Config]:
        """
        Parses an INI file into a configuration

        This method skips the processing of environment variables and command-line parameters.

        Args:
            ini_file_path: Path to an INI file

        Returns:
            A parsed configuration or an error
        """
        processor = cls.processor_()
        return processor.process_ini_file(ini_file_path)

    @classmethod
    def parse_command_line_(
        cls: typing.Type[_Config],
        cwd: typing.Optional[Path] = None,
        args: typing.Optional[typing.Sequence[str]] = None,
        env: typing.Optional[typing.Mapping[str, str]] = None,
    ) -> Res[typing.Union[_Config, SpecialAction]]:
        """
        Parses multiple information sources, returns a configuration, a command or an error

        Default values are taken from the current working directory, the script command line
        arguments, and the current environment variables.

        Args:
            cwd: Directory used as a base for the configuration file relative paths
            args: Command line arguments
            env: Environment variables

        Returns:
            A parsed configuration or an error
        """
        if cwd is None:
            cwd = Path.cwd()
        if args is None:
            args = sys.argv[1:]
        if env is None:
            env = os.environ
        processor = cls.processor_()
        return processor.process_command_line(cwd, args, env)

    @classmethod
    def from_command_line_(
        cls: typing.Type[_Config],
        cwd: typing.Optional[Path] = None,
        args: typing.Optional[typing.Sequence[str]] = None,
        env: typing.Optional[typing.Mapping[str, str]] = None,
    ) -> _Config:
        """
        Parses multiple information sources into a configuration and display help on error

        Default values are taken from the current working directory, the script command line
        arguments, and the current environment variables.

        Args:
            cwd: Directory used as a base for the configuration file relative paths
            args: Command line arguments
            env: Environment variables

        Returns:
            A parsed configuration
        """
        if cwd is None:
            cwd = Path.cwd()
        if args is None:
            args = sys.argv[1:]
        if env is None:
            env = os.environ
        res = cls.parse_command_line_(cwd, args, env)

        if isinstance(res, cls):
            return res

        if isinstance(res, Err):
            print("Encountered errors:")
            res.pretty_print()
            print(" ")
            cls.get_argument_parser_().print_help()
            sys.exit(1)

        assert isinstance(res, SpecialAction)
        if res == SpecialAction.HELP:
            cls.processor_().argument_parser.print_help()
            sys.exit(0)
        elif res == SpecialAction.VERSION:
            v = cls.version_()
            if v is None:
                v = "Unknown version number"
            print(v)
            sys.exit(0)
        else:
            raise NotImplementedError(f"Unknown special action {res}")

    # endregion

    # region Config: argparse fallback for documentation purposes

    @classmethod
    def get_argument_parser_(cls: typing.Type[_Config]) -> argparse.ArgumentParser:
        """
        Returns an :class:`argparse.ArgumentParser` for documentation purposes

        This may be removed or deprecated in later versions of configpile, if we write our
        own help/usage display function, and propose a Sphinx extension.
        """
        return cls.processor_().argument_parser

    # endregion



"""
Arguments processing

This module contains the internal machinery that processes environment variables, configuration
files and command-line parameters.

As of March 22, 2022, configpile is still pretty much influenced by :mod:`argparse`, and the
machinery below supports a subset of :mod:`argparse` functionality. Later on, we may cut ties
with :mod:`argparse`, add our own help/usage message writing, our own Sphinx extension and
encourage extending those processing classes.

.. rubric:: Types

This module uses the following types.

.. py:data:: _Config

    Type of the configuration dataclass to construct
"""

from __future__ import annotations

import argparse
import configparser
import sys
import warnings
from configparser import ConfigParser
from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from typing_extensions import Annotated, get_args, get_origin, get_type_hints

from .arg import Arg, Expander, Param
from .enums import SpecialAction
from .handlers import CLHandler, CLPos, CLSpecialAction, CLStdHandler, KVHandler
from .userr import Err, Res
from .util import ClassDoc, filter_types_single

if TYPE_CHECKING:
    from .config import Config

_Config = TypeVar("_Config", bound="Config")


@dataclass
class State:
    """
    Describes the (mutable) state of a configuration being parsed
    """

    instances: Dict[str, List[Any]]  #: Contains the sequence of values for each parameter
    config_files_to_process: List[Path]  #: Contains a list of configuration files to process
    special_action: Optional[SpecialAction]  #: Contains a special action if flag was encountered

    def append(self, key: str, value: Any) -> None:
        """
        Appends a value to a parameter

        No type checking is performed, be careful.

        Args:
            key: Parameter name
            value: Value to append
        """
        assert key in self.instances, f"{key} is not a Param name"
        self.instances[key] = [*self.instances[key], value]

    @staticmethod
    def make(params: Iterable[Param[Any]]) -> State:
        """
        Creates the initial state, populated with the default values when present

        Args:
            params: Sequence of parameters

        Raises:
            ValueError: If a default value cannot be parsed correctly

        Returns:
            The initial mutable state
        """
        instances: Dict[str, List[Any]] = {}

        for p in params:
            assert p.name is not None, "Arguments have names after initialization"
            if p.default_value is not None:
                res = p.parser.parse(p.default_value)
                if isinstance(res, Err):
                    raise ValueError(f"Invalid default {p.default_value} for parameter {p.name}")
                instances[p.name] = [res]
            else:
                instances[p.name] = []
        return State(instances, config_files_to_process=[], special_action=None)


@dataclass(frozen=True)
class IniProcessor:
    """
    INI configuration file processor
    """

    section_strict: Mapping[str, bool]  #: Sections and their strictness
    kv_handlers: Mapping[str, KVHandler]  #: Handler for key/value pairs

    def _process(self, parser: ConfigParser, state: State) -> Sequence[Err]:
        """
        Processes a filled ConfigParser

        Args:
            parser: INI data to parse
            state: Mutable state to update

        Returns:
            Errors that occurred, if any
        """
        errors: List[Err] = []
        try:
            for section_name in parser.sections():
                if section_name in self.section_strict:
                    for key, value in parser[section_name].items():
                        err: Optional[Err] = None
                        if key in self.kv_handlers:
                            res = self.kv_handlers[key].handle(value, state)
                            if isinstance(res, Err):
                                err = res
                        else:
                            if self.section_strict[section_name]:
                                err = Err.make(f"Unknown key {key}")
                        if err is not None:
                            errors.append(err.in_context(ini_section=section_name))
        except configparser.Error:
            errors.append(Err.make("Parse error"))
        except IOError:
            errors.append(Err.make("IO Error"))
        return errors

    def process_string(self, ini_contents: str, state: State) -> Optional[Err]:
        """
        Processes a configuration file given as a string

        Args:
            ini_contents: Contents of the INI file
            state: Mutable state to update

        Returns:
            An optional error
        """
        errors: List[Err] = []
        parser = ConfigParser()
        try:
            parser.read_string(ini_contents)
            errors.extend(self._process(parser, state))
        except configparser.Error:
            errors.append(Err.make("Parse error"))
        except IOError:
            errors.append(Err.make("IO Error"))
        if errors:
            return Err.collect(*errors)
        else:
            return None

    def process(self, ini_file_path: Path, state: State) -> Optional[Err]:
        """
        Processes a configuration file

        Args:
            ini_path: Path to the INI file
            state: Mutable state to update

        Returns:
            An optional error
        """
        errors: List[Err] = []
        if not ini_file_path.exists():
            return Err.make(f"Config file {ini_file_path} does not exist")
        if not ini_file_path.is_file():
            return Err.make(f"Path {ini_file_path} is not a file")
        parser = ConfigParser()
        parser.optionxform = str  # disable conversion to lower-case
        try:
            with open(ini_file_path, "r", encoding="utf-8") as file:
                parser.read_file(file)
                errors.extend(self._process(parser, state))
        except configparser.Error:
            errors.append(Err.make("Parse error"))
        except IOError:
            errors.append(Err.make("IO Error"))
        if errors:
            return Err.collect(*errors)
        else:
            return None


@dataclass
class ProcessorFactory(Generic[_Config]):
    """
    Describes a processor in construction

    This factory is passed to the different arguments present in the configuration.
    """

    #: List of parameters indexed by their field name
    params_by_name: Dict[str, Param[Any]]

    #: Argument parser to update, used to display help and for the Sphinx documentation
    argument_parser: argparse.ArgumentParser

    #: Argument parser group for commands
    ap_commands: argparse._ArgumentGroup

    #: Argument parser group for required parameters
    ap_required: argparse._ArgumentGroup

    #: Argument parser group for optional parameters
    ap_optional: argparse._ArgumentGroup

    #: Handlers for environment variables
    env_handlers: Dict[str, KVHandler]  # = {}

    #: List of INI sections with their corresponding strictness
    ini_section_strict: Dict[str, bool]

    #: List of handlers for key/value pairs present in INI files
    ini_handlers: Dict[str, KVHandler]

    #: List of command line flag handlers
    cl_flag_handlers: Dict[str, CLHandler]

    #: List of positional arguments
    cl_positionals: List[Param[Any]]

    validators: List[Callable[[_Config], Optional[Err]]]

    @staticmethod
    def make(config_type: Type[_Config]) -> ProcessorFactory[_Config]:
        """
        Constructs an empty processor factory

        Args:
            config_type: Configuration to process

        Returns:
            A processor factory
        """
        # fill program name from script invocation
        prog = config_type.prog_
        if prog is None:
            prog = sys.argv[0]

        # fill description from class docstring
        description: Optional[str] = config_type.description_
        if description is None:
            description = config_type.__doc__

        argument_parser = argparse.ArgumentParser(
            prog=prog,
            description=description,
            formatter_class=argparse.RawTextHelpFormatter,
            add_help=False,
        )
        argument_parser._action_groups.pop()  # pylint: disable=protected-access
        return ProcessorFactory(
            params_by_name={},
            argument_parser=argument_parser,
            ap_commands=argument_parser.add_argument_group("commands"),
            ap_optional=argument_parser.add_argument_group("optional arguments"),
            ap_required=argument_parser.add_argument_group("required arguments"),
            env_handlers={},
            ini_section_strict={s.name: s.strict for s in config_type.ini_sections_()},
            ini_handlers={},
            cl_flag_handlers={},
            cl_positionals=[],
            validators=[*config_type.validators_()],
        )


@dataclass(frozen=True)
class Processor(Generic[_Config]):
    """
    Configuration processor
    """

    #: Configuration to parse
    config_type: Type[_Config]

    #: Completed argument parser, used only for documentation purposes (CLI and Sphinx)
    argument_parser: argparse.ArgumentParser

    #: Environment variable handlers
    env_handlers: Mapping[str, KVHandler]

    #: INI file processor
    ini_processor: IniProcessor

    #: Command line arguments handler
    cl_handler: CLStdHandler

    #: Dictionary of parameters by field name
    params_by_name: Mapping[str, Param[Any]]

    validators: Sequence[Callable[[_Config], Optional[Err]]]

    @staticmethod
    def process_fields(config_type: Type[_Config]) -> Sequence[Arg]:
        """
        Returns a sequence of the arguments present in a configuration, with updated data

        Args:
            config_type: Configuration to process

        Returns:
            Sequence of arguments
        """
        args: List[Arg] = []
        docs: ClassDoc[_Config] = ClassDoc.make(config_type)
        th = get_type_hints(config_type, include_extras=True)
        for name, typ in th.items():
            arg: Optional[Arg] = None
            if get_origin(typ) is ClassVar:
                a = getattr(config_type, name)
                if isinstance(a, Arg):
                    assert isinstance(a, Expander), "Only commands (Cmd) can be class attributes"
                    arg = a
            if get_origin(typ) is Annotated:
                param = filter_types_single(Param, get_args(typ))
                if param is not None:
                    arg = param
            if arg is not None:
                help_lines = docs[name]
                if help_lines is None:
                    hlp = ""
                else:
                    hlp = "\n".join(help_lines)
                arg = arg.updated(name, hlp, config_type.env_prefix_)
                args.append(arg)
        return args

    @staticmethod
    def make(
        config_type: Type[_Config],
    ) -> Processor[_Config]:
        """
        Creates the processor corresponding to a configuration
        """

        pf = ProcessorFactory.make(config_type)
        for arg in Processor.process_fields(config_type):
            arg.update_processor(pf)

        # if these flags are no longer provided by default, update the overview concept notebook
        # in the documentation
        pf.cl_flag_handlers["-h"] = CLSpecialAction(SpecialAction.HELP)
        pf.cl_flag_handlers["--help"] = CLSpecialAction(SpecialAction.HELP)

        return Processor(
            config_type=config_type,
            argument_parser=pf.argument_parser,
            env_handlers=pf.env_handlers,
            ini_processor=IniProcessor(pf.ini_section_strict, pf.ini_handlers),
            cl_handler=CLStdHandler(pf.cl_flag_handlers, CLPos(pf.cl_positionals)),
            params_by_name=pf.params_by_name,
            validators=pf.validators,
        )

    def _process_config(self, cwd: Path, state: State) -> Optional[Err]:
        """
        Processes configuration files if such processing was requested by a handler

        Args:
            cwd: Working directory, base path for relative paths of config files
            state: Mutable state to update

        Returns:
            An optional error
        """
        paths = state.config_files_to_process
        state.config_files_to_process = []
        errors: List[Err] = []
        for p in paths:
            err = self.ini_processor.process(cwd / p, state)
            if err is not None:
                errors.append(err.in_context(ini_file=p))
        return Err.collect(*errors)

    def process_ini_contents(self, ini_contents: str) -> Res[_Config]:
        """
        Processes the configuration in INI format contained in a string

        Args:
            ini_contents: Multiline string containing information in INI format

        Returns:
            The parsed configuration or an error
        """
        state = self._state_with_default_values()
        err = self.ini_processor.process_string(ini_contents, state)
        if err is not None:
            return err
        return self._finish_processing_state(state)

    def process_ini_file(self, ini_file_path: Path) -> Res[_Config]:
        """
        Processes the configuration contained in an INI file

        Args:
            ini_file_path: Path to the file to parse

        Returns:
            The parsed configuration or an error
        """
        state = self._state_with_default_values()
        err = self.ini_processor.process(ini_file_path, state)
        if err is not None:
            return err
        return self._finish_processing_state(state)

    def process(
        self,
        cwd: Path,
        args: Sequence[str],
        env: Mapping[str, str],
    ) -> Res[Union[_Config, SpecialAction]]:
        """
        Processes command-line arguments (deprecated)

        See also: :meth:`.process_command_line`
        """
        warnings.warn(
            "process has been deprecated, use process_command_line instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.process_command_line(cwd, args, env)

    def _finish_processing_state(self, state: State) -> Res[_Config]:
        """
        Finishes the processing of the state

        This method performs:

        - the collection of parameter values
        - the final validation using validation methods

        Args:
            state: State after parsing all elements, must have :attr:`.State.special_action` set
                   to :data:`None`

        Returns:
            The parsed configuration or an error
        """
        assert state.special_action is None
        errors: List[Err] = []
        collected: Dict[str, Any] = {}
        for name, param in self.params_by_name.items():
            instances = state.instances[name]
            res = param.collector.collect(instances)
            if isinstance(res, Err):
                errors.append(res.in_context(param=name))
            else:
                collected[name] = res

        if errors:
            return Err.collect1(*errors)
        c: _Config = self.config_type(**collected)
        validation_error: Optional[Err] = Err.collect(*[f(c) for f in self.validators])
        if validation_error is not None:
            return validation_error
        else:
            return c

    def _state_with_default_values(self) -> State:
        """
        Returns a new state instance with default values populated
        """
        return State.make(self.params_by_name.values())

    def process_command_line(
        self,
        cwd: Path,
        args: Sequence[str],
        env: Mapping[str, str],
    ) -> Res[Union[_Config, SpecialAction]]:
        """
        Processes command-line arguments, configuration files and environment variables

        Args:
            cwd: Working directory, used as a base for configuration file relative paths
            args: Command line arguments to parse
            env: Environment variables

        Returns:
            Either a parsed configuration, a special action to execute, or (a list of) errors
        """
        errors: List[Err] = []
        state = self._state_with_default_values()
        # process environment variables
        for key, value in env.items():
            handler = self.env_handlers.get(key)
            if handler is not None:
                err = handler.handle(value, state)
                if err is not None:
                    errors.append(err.in_context(environment_variable=key))
            err = self._process_config(cwd, state)
            if err is not None:
                errors.append(err.in_context(environment_variable=key))
        # process command line arguments
        rest_args: Sequence[str] = args
        while rest_args:
            rest_args, err = self.cl_handler.handle(rest_args, state)
            if err is not None:
                errors.append(err)
            err = self._process_config(cwd, state)
            if err is not None:
                errors.append(err)

        if state.special_action is not None:
            return state.special_action

        if errors:
            return Err.collect1(*errors)

        return self._finish_processing_state(state)



"""
Value consolidation

This module defines :class:`.Collector`, which is a class that collects a sequence of parsed values
into a single value.

For example, one may want that the last value for a given parameter overwrites the previous values
(:meth:`.Collector.keep_last`), or that all values are returned in a sequence
(:meth:`.Collector.append`, :meth:`.Collector.append1`).

.. rubric:: Types

This module uses the following types.

.. py:data:: _Value

    Type of the values to collect

.. py:data:: _Item

    Item in a sequence

"""

from __future__ import annotations

import abc
from typing import Any, Generic, List, Mapping, Sequence, TypeVar

from .userr import Err, Res

_Value = TypeVar("_Value")

_Item = TypeVar("_Item")


class Collector(abc.ABC, Generic[_Value]):
    """
    Collects argument instances and computes the final value
    """

    @abc.abstractmethod
    def arg_required(self) -> bool:
        """
        Returns whether one instance of the argument needs to be present
        """

    @abc.abstractmethod
    def collect(self, seq: Sequence[_Value]) -> Res[_Value]:
        """
        Collects a sequence of values into a result

        Args:
            seq: Sequence of parsed values

        Returns:
            Either the consolidated value or an error
        """

    @abc.abstractmethod
    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        """
        Returns the arguments using in documentation (piggy backing on argparse)
        """

    @staticmethod
    def keep_last() -> Collector[_Value]:
        """
        Returns a collector that keeps the last value
        """

        class _KeepLast(Collector[_Value]):
            def arg_required(self) -> bool:
                return True

            def collect(self, seq: Sequence[_Value]) -> Res[_Value]:
                if not seq:  # no instances provided
                    return Err.make("Argument is required")
                else:  # instances are provided
                    return seq[-1]

            def argparse_argument_kwargs(self) -> Mapping[str, Any]:
                return {"action": "store"}

        return _KeepLast()

    @staticmethod
    def append() -> Collector[Sequence[_Item]]:
        """
        Returns a collector that appends sequences
        """

        class _Append(Collector[Sequence[_Item]]):
            def arg_required(self) -> bool:
                return False

            def collect(self, seq: Sequence[Sequence[_Item]]) -> Res[Sequence[_Item]]:
                res: List[_Item] = []
                for i in seq:
                    res.extend(i)
                return res

            def argparse_argument_kwargs(self) -> Mapping[str, Any]:
                return {"action": "append"}

        return _Append()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['configpile']

package_data = \
{'': ['*']}

install_requires = \
['class-doc>=0.2.6,<0.3.0', 'typing-extensions>=4.1.1,<5.0.0']

extras_require = \
{'docs': ['myst-nb>=0.13.2,<0.14.0',
          'sphinx==4.3.2',
          'sphinx-argparse>=0.3.1,<0.4.0',
          'sphinx-autodoc-typehints>=1.17.0,<2.0.0',
          'sphinx-book-theme>=0.2.0,<0.3.0',
          'sphinxcontrib-programoutput>=0.17,<0.18'],
 'parsy': ['parsy>=1.4.0,<2.0.0'],
 'rich': ['rich>=11.2.0,<12.0.0']}

setup_kwargs = {
    'name': 'configpile',
    'version': '10.0.3',
    'description': 'Configuration from command line parameters, configuration files and environment variables',
    'long_description': 'ConfigPile\n==========\n\nA configuration library for Python\n\nSee the website https://denisrosset.github.io/configpile \n',
    'author': 'Denis Rosset',
    'author_email': 'physics@denisrosset.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/denisrosset/configpile.git',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
This module defines the handlers that are used during processing.

.. rubric:: Types

This module uses the following types.

.. py:data:: _Value

    Value being parsed by a :class:`~configpile.parsers.Parser`
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, List, Mapping, Optional, Sequence, Tuple, TypeVar

from .arg import Positional
from .enums import SpecialAction
from .userr import Err, in_context

if TYPE_CHECKING:
    from .arg import Expander, Param
    from .processor import State

_Value = TypeVar("_Value")


class CLHandler(ABC):
    """
    A handler for command-line arguments
    """

    @abstractmethod
    def handle(self, args: Sequence[str], state: State) -> Tuple[Sequence[str], Optional[Err]]:
        """
        Processes arguments, possibly updating the state or returning errors

        Args:
            args: Command-line arguments not processed yet
            state: (Mutable) state to possibly update

        Returns:
            The updated command-line and an optional error
        """


@dataclass(frozen=True)
class CLSpecialAction(CLHandler):
    """
    A handler that sets the special action
    """

    special_action: SpecialAction  #: Special action to set

    def handle(self, args: Sequence[str], state: State) -> Tuple[Sequence[str], Optional[Err]]:
        if state.special_action is not None:
            before = state.special_action.name
            now = self.special_action.name
            err = Err.make(f"We had already action {before}, conflicts with action {now}")
            return (args, err)
        state.special_action = self.special_action
        return (args, None)


@dataclass(frozen=True)
class CLInserter(CLHandler):
    """
    Handler that expands a flag into a sequence of args inserted into the command line to be parsed
    """

    #: Arguments inserted in the command-line
    inserted_args: Sequence[str]

    def handle(self, args: Sequence[str], state: State) -> Tuple[Sequence[str], Optional[Err]]:
        return ([*self.inserted_args, *args], None)


@dataclass(frozen=True)
class CLParam(CLHandler, Generic[_Value]):
    """
    Parameter handler

    Takes a single string argument from the command line, parses it and pushes into the
    corresponding sequence of instances
    """

    #: Parameter to handle
    param: Param[_Value]

    def action(self, value: _Value, state: State) -> Optional[Err]:
        """
        A method called on the successful parse of a value

        Can be overridden. By default does nothing.

        Args:
            value: Parsed value
            state: State to possibly update

        Returns:
            An optional error
        """
        return None

    def handle(self, args: Sequence[str], state: State) -> Tuple[Sequence[str], Optional[Err]]:
        if args:
            res = self.param.parser.parse(args[0])
            if isinstance(res, Err):
                return (args[1:], res.in_context(param=self.param.name))
            else:
                assert self.param.name is not None, "Names are assigned after initialization"
                err = in_context(self.action(res, state), param=self.param.name)
                state.instances[self.param.name] = [*state.instances[self.param.name], res]
                return (args[1:], err)
        else:
            return (
                args,
                Err.make("Expected value, but no argument present", param=self.param.name),
            )


@dataclass(frozen=True)
class CLConfigParam(CLParam[Sequence[Path]]):
    """
    A configuration file parameter handler

    If paths are successfully parsed, it appends configuration files to be parsed to the current
    state.
    """

    def action(self, value: Sequence[Path], state: State) -> Optional[Err]:
        state.config_files_to_process.extend(value)
        return None


@dataclass
class CLPos(CLHandler):
    """
    Handles positional parameters

    Note that this handler has state, namely the positional parameters that are still expected.
    """

    pos: List[Param[Any]]  #: (Mutable) list of positional parameters

    @staticmethod
    def make(seq: Sequence[Param[Any]]) -> CLPos:
        """
        Constructs a positional parameter handler from a sequence of positional parameters

        Args:
            seq: Positional parameters

        Returns:
            Handler
        """
        assert all([p.positional is not None for p in seq]), "All parameters should be positional"
        assert all(
            [not p.positional.should_be_last() for p in seq[:-1] if p.positional is not None]
        ), "Positional parameters with a variable number of arguments should be last"
        l = list(seq)  # makes a mutable copy
        return CLPos(l)

    def handle(self, args: Sequence[str], state: State) -> Tuple[Sequence[str], Optional[Err]]:
        if not args:
            return (args, None)  # should not happen ,but let's not crash
        if not self.pos:
            return (args[1:], Err.make(f"Unknown argument {args[0]}"))
        p = self.pos[0]
        assert p.name is not None
        res = p.parser.parse(args[0])
        if isinstance(res, Err):
            return (args[1:], in_context(res, param=p.name))
        else:
            state.append(p.name, res)
            if p.positional == Positional.ONCE:
                self.pos = self.pos[1:]
            return (args[1:], None)


@dataclass(frozen=True)
class CLStdHandler(CLHandler):
    """
    The standard command line arguments handler

    It processes arguments one by one. If it recognizes a flag, the corresponding handler is
    called. Otherwise, control is passed to the fallback handler, which by default processes
    positional parameters.
    """

    flags: Mapping[str, CLHandler]
    fallback: CLHandler

    def handle(self, args: Sequence[str], state: State) -> Tuple[Sequence[str], Optional[Err]]:
        if not args:
            return (args, None)
        flag = args[0]
        handler = self.flags.get(flag)
        if handler is not None:
            next_args, err = handler.handle(args[1:], state)
            err = in_context(err, flag=flag)
            return next_args, err
        else:
            return self.fallback.handle(args, state)


class KVHandler(ABC):
    """
    Handler for key/value pairs found for example in environment variables or INI files

    Note that the key is not stored/processed in this class.
    """

    @abstractmethod
    def handle(self, value: str, state: State) -> Optional[Err]:
        """
        Processes

        Args:
            value: Value to parse and process
            state: State to update

        Returns:
            An error if an error occurred
        """


@dataclass(frozen=True)
class KVParam(KVHandler, Generic[_Value]):
    """
    Handler for the value following a key corresponding to a parameter
    """

    #: Parameter to handle
    param: Param[_Value]

    def action(self, value: _Value, state: State) -> Optional[Err]:
        """
        A method called on the successful parse of a value

        Can be overridden. By default does nothing.

        Args:
            value: Parsed value
            state: State to possibly update

        Returns:
            An optional error
        """
        return None

    def handle(self, value: str, state: State) -> Optional[Err]:
        res = self.param.parser.parse(value)
        if isinstance(res, Err):
            return res
        else:
            assert self.param.name is not None
            err = self.action(res, state)
            state.instances[self.param.name] = [*state.instances[self.param.name], res]
            return in_context(err, param=self.param.name)


@dataclass(frozen=True)
class KVConfigParam(KVParam[Sequence[Path]]):
    """
    Handler for the configuration file value following a key corresponding to a parameter
    """

    def action(self, value: Sequence[Path], state: State) -> Optional[Err]:
        state.config_files_to_process.extend(value)
        return None



"""
This module defines enumerations that are used by the library.

We collect them in this module to solve the circular import problem.
"""

from enum import Enum


class Derived(Enum):
    """
    Describes automatic handling of an argument name
    """

    #: Derives the name/key using snake_case
    SNAKE_CASE = 1

    #: Derives the name/key using SNAKE_CASE (and sets uppercase, default for env. variables)
    SNAKE_CASE_UPPER_CASE = 2

    #: Derives the argument name using kebab-case (default for INI file keys and cmd line flags)
    KEBAB_CASE = 3

    def derive(self, name: str) -> str:
        """
        Returns a derived name from a field name

        Args:
            name: Field name in Python identifier format
        """
        assert str.isidentifier(name)
        if self == Derived.SNAKE_CASE:
            return name
        elif self == Derived.SNAKE_CASE_UPPER_CASE:
            return name.upper()
        elif self == Derived.KEBAB_CASE:
            return name.replace("_", "-")
        else:
            raise NotImplementedError


class ForceCase(Enum):
    """
    Describes whether a string is normalized to lower or upper case before processing
    """

    NO_CHANGE = 0  #: Keep case
    UPPER = 1  #: Change to uppercase
    LOWER = 2  #: Change to lowercase


class Positional(Enum):
    """
    Describes the positional behavior of a parameter
    """

    ONCE = 1  #: Eats a single positional value
    ZERO_OR_MORE = 2  #: Eats as many positional values as possible
    ONE_OR_MORE = 3  #: Eats as many positional values as possible, but at least one

    def should_be_last(self) -> bool:
        """
        Returns whether a positional parameter should be the last one
        """
        return self in {Positional.ZERO_OR_MORE, Positional.ONE_OR_MORE}


class SpecialAction(Enum):
    """
    Describes special actions that do not correspond to normal execution
    """

    HELP = "help"  #: Display a help message
    VERSION = "version"  #: Print the version number



__version__ = "10.0.3"

from . import parsers
from .arg import Derived, Expander, Param, Positional
from .config import Config
from .parsers import ForceCase, Parser
from .userr import Err, Res

# when updating this, update the docs/source/api.rst list

__all__ = [
    "Config",
    "Derived",
    "Err",
    "Expander",
    "Res",
    "Param",
    "Parser",
    "Positional",
    "parsers",
    "ForceCase",
    "fun",
]
